# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from dataclasses import dataclass
from typing import List

import jax
import jax.numpy as jnp
import numpy as np
from jax._src import dtypes
from jax.sharding import Mesh, NamedSharding, PartitionSpec

import tpu_inference.kernels.mla.v1.kernel as mla
import tpu_inference.kernels.ragged_paged_attention.v3.kernel as rpa
import tpu_inference.kernels.ragged_paged_attention.v3.kernel_hd64 as rpa_hd64
from tpu_inference import utils
from tpu_inference.layers.common.sharding import ShardingAxisName
from tpu_inference.logger import init_logger
from tpu_inference.utils import to_jax_dtype

logger = init_logger(__name__)

DEFAULT_KV_CACHE_DTYPE = jnp.bfloat16


@dataclass
class KVCacheMetadata:
    """
    Used to store metadata about the KV cache for logging in the KV cache manager.
    Specifcally, with Hybrid KV cache, we can have multiple KV cache types
    so we need to store the metadata for each KV cache type separately
    """
    count: int = 0
    shape: tuple = None
    dtype: jnp.dtype = None
    sharding: NamedSharding = None


def get_kv_cache_shape_with_mesh(mesh: Mesh,
                                 total_num_pages: int,
                                 page_size: int,
                                 actual_num_kv_heads: int,
                                 actual_head_dim: int,
                                 kv_dtype: any,
                                 use_mla: bool = False):
    """Gets the KV cache shape based on the mesh configuration."""

    axis_name = ShardingAxisName.ATTN_HEAD
    model_cnt = utils.get_mesh_shape_product(mesh, axis_name)

    # NOTE(chengjiyao): Currently, the attention kernel is tailored to the
    # specific model, rather than being determined by the head_dim. If new
    # models are introduced with a head_dim of 64, this will require additional
    # model-specific adjustments.
    if use_mla:
        # No assertion needed: MLA compresses all KV into a single latent vector,
        # so actual_num_kv_heads is never used in mla.get_kv_cache_shape().
        get_kv_cache_shape_fn = mla.get_kv_cache_shape
        shape = list(
            get_kv_cache_shape_fn(total_num_pages, page_size, actual_head_dim,
                                  kv_dtype))
    else:
        assert actual_num_kv_heads % model_cnt == 0
        get_kv_cache_shape_fn = (
            rpa_hd64.get_kv_cache_shape if actual_head_dim == 64 \
                else rpa.get_kv_cache_shape
        )
        shape = list(
            get_kv_cache_shape_fn(total_num_pages, page_size,
                                  actual_num_kv_heads // model_cnt,
                                  actual_head_dim, kv_dtype))
        shape[2] *= model_cnt
    return tuple(shape)


def create_kv_caches(
    num_blocks: int,
    block_size: int,
    num_kv_heads: int,
    head_size: int,
    mesh: Mesh,
    layer_names: List[str],
    cache_dtype: jnp.dtype = DEFAULT_KV_CACHE_DTYPE,
    use_mla: bool = False,
) -> List[jax.Array]:
    """
    Creates a list of KV cache where each array mapps to single attention layer.

    The shape of the KV cache per layer is:
    (num_blocks, block_size, cdiv(num_kv_heads * 2, packing), packing, head_dim)
    where packing = (32 // dtype bits)

    Args:
        num_blocks: The number of blocks in the KV cache.
        block_size: The size of each block in the KV cache.
        num_kv_heads: The number of KV heads in the KV cache.
        head_size: The size of each head in the KV cache.
        mesh: The mesh to shard the KV caches across.
        layer_names: The names of the decoder layers in the model.
        cache_dtype: The datatype of KV cache.

    Returns:
        A list of KV caches, one per each decoder layer in the model.

    """
    # TODO(xiang): fix this together with get_kv_cache_spec
    # cache_dtype = kv_cache_spec.dtype

    cache_shape = get_kv_cache_shape_with_mesh(mesh, num_blocks, block_size,
                                               num_kv_heads, head_size,
                                               cache_dtype, use_mla)

    if use_mla:
        sharding = NamedSharding(mesh,
                                 PartitionSpec(ShardingAxisName.MLP_TENSOR))
    else:
        sharding = NamedSharding(
            mesh,
            PartitionSpec(ShardingAxisName.ATTN_DATA, None,
                          ShardingAxisName.ATTN_HEAD))

    def _allocate() -> jax.Array:
        return jnp.empty(
            shape=cache_shape,
            dtype=cache_dtype,
        )

    sharded_allocate = jax.jit(_allocate, out_shardings=sharding)
    kv_caches = []
    for _ in layer_names:
        kv_caches.append(sharded_allocate())
    return kv_caches


def get_attention_page_size_bytes(mesh, block_size, num_kv_heads, head_size,
                                  dtype, use_mla) -> int:
    jax_dtype = to_jax_dtype(dtype)
    bits = dtypes.itemsize_bits(jax_dtype)
    kv_cache_shape = get_kv_cache_shape_with_mesh(
        mesh=mesh,
        total_num_pages=1,
        page_size=block_size,
        actual_num_kv_heads=num_kv_heads,
        actual_head_dim=head_size,
        kv_dtype=jax_dtype,
        use_mla=use_mla,
    )
    return (bits * np.prod(kv_cache_shape)) // 8



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import os
import time
from dataclasses import dataclass, field

import pytest
from vllm import LLM, SamplingParams


@dataclass
class TestConfig:
    """Configuration for TP test runs."""
    max_model_len: int = 512
    max_num_batched_tokens: int = 128
    max_num_seqs: int = 16
    num_prompts: int = 16

    @classmethod
    def for_performance(cls) -> "TestConfig":
        return cls(
            max_model_len=1024,
            max_num_batched_tokens=2048,
            max_num_seqs=2048,
            num_prompts=2048,
        )


@dataclass
class InferenceConfig:
    """Configuration for a single inference run."""
    model_name: str
    tensor_parallel_size: int = 1
    pipeline_parallel_size: int = 1
    max_model_len: int = 512
    max_num_batched_tokens: int = 128
    max_num_seqs: int = 16
    additional_config: dict = field(default_factory=dict)
    gpu_memory_utilization: float = 0.80
    kv_cache_dtype: str = "auto"
    enable_prefix_caching: bool = False


def generate_test_prompts(num_prompts: int = 256) -> list[str]:
    base_text = (
        "The rapid advancement of artificial intelligence has transformed "
        "numerous industries and continues to reshape our understanding of "
        "technology's potential. Machine learning algorithms have become "
        "increasingly sophisticated, enabling computers to perform tasks "
        "that were once thought to require human intelligence. From natural "
        "language processing to computer vision, AI systems are now capable "
        "of understanding context, recognizing patterns, and making decisions "
        "with remarkable accuracy. ")
    return [
        f"Prompt {i}: {base_text} What are your thoughts on this topic?"
        for i in range(num_prompts)
    ]


@pytest.fixture
def sampling_params():
    return SamplingParams(
        temperature=0.0,
        max_tokens=32,
        ignore_eos=True,
        logprobs=1,
    )


def _run_inference(
    config: InferenceConfig,
    test_prompts: list[str],
    sampling_params: SamplingParams,
) -> tuple[list, float]:
    """Run inference with the given configuration."""
    llm = LLM(
        model=config.model_name,
        max_model_len=config.max_model_len,
        tensor_parallel_size=config.tensor_parallel_size,
        pipeline_parallel_size=config.pipeline_parallel_size,
        gpu_memory_utilization=config.gpu_memory_utilization,
        max_num_batched_tokens=config.max_num_batched_tokens,
        max_num_seqs=config.max_num_seqs,
        additional_config=config.additional_config,
        kv_cache_dtype=config.kv_cache_dtype,
        enable_prefix_caching=config.enable_prefix_caching,
    )

    start_time = time.time()
    outputs = llm.generate(test_prompts, sampling_params)
    elapsed_time = time.time() - start_time

    del llm
    time.sleep(10)
    return outputs, elapsed_time


def _check_performance(
    test_name: str,
    baseline_time: float,
    tp_time: float,
    num_prompts: int,
    min_speedup: float,
):
    """Verify tensor parallelism provides expected speedup."""
    speedup = baseline_time / tp_time if tp_time > 0 else 0

    print(f"✓ {test_name} performance test results:")
    print(f"  Number of prompts: {num_prompts}")
    print(f"  Baseline time: {baseline_time:.2f}s")
    print(f"  Tensor parallel time: {tp_time:.2f}s")
    print(f"  Speedup: {speedup:.2f}x")
    print(f"  Baseline throughput: {num_prompts/baseline_time:.2f} prompts/s")
    print(f"  Tensor parallel throughput: {num_prompts/tp_time:.2f} prompts/s")

    assert speedup >= min_speedup, (
        f"Tensor parallelism did not provide expected speedup "
        f"({min_speedup:.2f}x): {speedup:.2f}x")


def _test_tensor_parallelism_performance(
    sampling_params: SamplingParams,
    model_name: str,
    tensor_parallel_size: int,
    pipeline_parallel_size: int = 1,
    additional_config: dict | None = None,
    min_speedup: float = 1.05,
):
    """Performance test for tensor parallelism."""
    cfg = TestConfig.for_performance()
    test_prompts = generate_test_prompts(cfg.num_prompts)

    tp_config = InferenceConfig(
        model_name=model_name,
        tensor_parallel_size=tensor_parallel_size,
        pipeline_parallel_size=pipeline_parallel_size,
        max_model_len=cfg.max_model_len,
        max_num_batched_tokens=cfg.max_num_batched_tokens //
        tensor_parallel_size,
        max_num_seqs=cfg.max_num_seqs // tensor_parallel_size,
        additional_config=additional_config or {},
    )
    _, tp_time = _run_inference(tp_config, test_prompts, sampling_params)

    baseline_config = InferenceConfig(
        model_name=model_name,
        tensor_parallel_size=1,
        pipeline_parallel_size=1,
        max_model_len=cfg.max_model_len,
        max_num_batched_tokens=cfg.max_num_batched_tokens,
        max_num_seqs=cfg.max_num_seqs,
    )
    _, baseline_time = _run_inference(baseline_config, test_prompts,
                                      sampling_params)

    _check_performance(
        "Tensor parallelism",
        baseline_time,
        tp_time,
        len(test_prompts),
        min_speedup=min_speedup,
    )


def test_tp_performance(sampling_params: SamplingParams):
    """Performance test for tensor parallelism on vLLM models."""
    os.environ['MODEL_IMPL_TYPE'] = 'vllm'
    os.environ['SKIP_JAX_PRECOMPILE'] = '0'
    os.environ['VLLM_XLA_CHECK_RECOMPILATION'] = '1'

    _test_tensor_parallelism_performance(
        sampling_params=sampling_params,
        model_name="meta-llama/Llama-3.1-8B-Instruct",
        tensor_parallel_size=2,
        pipeline_parallel_size=1,
        min_speedup=1.05,
    )



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torchax
from jax.sharding import Mesh, PartitionSpec
from vllm.config import VllmConfig
from vllm.logger import init_logger
from vllm.model_executor.layers.fused_moe import FusedMoE, FusedMoEConfig
# yapf: disable
from vllm.model_executor.layers.linear import (ColumnParallelLinear,
                                               LinearBase,
                                               MergedColumnParallelLinear,
                                               QKVParallelLinear,
                                               ReplicatedLinear,
                                               RowParallelLinear)

from tpu_inference.layers.common.process_weights.linear_weights import \
    get_model_matmul_fusion_assignment
from tpu_inference.layers.common.quantization.configs import QuantLinearConfig
from tpu_inference.layers.common.sharding import ShardingAxisName
from tpu_inference.utils import TPU_SECOND_LAST_MINOR, get_mesh_shape_product

# yapf: enable

P = PartitionSpec

logger = init_logger(__name__)


class VllmQuantLinearConfig(QuantLinearConfig):

    def __init__(self, vllm_config: VllmConfig, mesh: Mesh, layer: LinearBase):
        assert isinstance(layer, LinearBase)

        super().__init__(
            enable_sp=vllm_config.compilation_config.pass_config.enable_sp,
            output_sizes=[layer.output_size])
        self.mesh = mesh
        self.tp_size = get_mesh_shape_product(self.mesh,
                                              ShardingAxisName.MLP_TENSOR)

        self.n_shards = get_mesh_shape_product(self.mesh,
                                               self.weight_sharding[0])

        if isinstance(layer, RowParallelLinear):
            self.weight_sharding = P(None, ShardingAxisName.ATTN_HEAD)
            if self.enable_sp:
                self.output_sharding = P(ShardingAxisName.MLP_TENSOR, None)
        elif isinstance(layer, ColumnParallelLinear):
            self.weight_sharding = P(ShardingAxisName.ATTN_HEAD, None)

            if self.enable_sp:
                self.input_sharding = P(ShardingAxisName.MLP_TENSOR, None)

            if isinstance(layer, MergedColumnParallelLinear) or isinstance(
                    layer, QKVParallelLinear):
                self.output_sizes = layer.output_sizes

            self.fuse_matmuls = get_model_matmul_fusion_assignment(
                vllm_config.model_config.model,
                vllm_config.scheduler_config.max_num_batched_tokens,
                vllm_config.parallel_config.tensor_parallel_size,
                layer._get_name())
        elif isinstance(layer, ReplicatedLinear):
            self.weight_sharding = P(None, None)
        else:
            logger.warning(
                "Unsupported linear layer type of %s. Can potentially yield "
                " bad performance.", type(layer))

        if isinstance(layer, QKVParallelLinear):
            self.num_proj = 3
        elif isinstance(layer, MergedColumnParallelLinear):
            self.num_proj = len(layer.output_sizes)
        else:
            self.num_proj = 1

        self.bias_sharding = P(self.weight_sharding[0])
        self.n_shards = get_mesh_shape_product(self.mesh,
                                               self.weight_sharding[0])

    def get_input_sharding(self, x: torchax.tensor.Tensor):
        if not self.enable_sp:
            return None
        token_num = x.shape[0]
        # NOTE(chengjiyao): make sure the sharded token_num is larger than TPU_SECOND_LAST_MINOR
        if token_num // self.tp_size < TPU_SECOND_LAST_MINOR:
            return None
        return self.input_sharding

    def get_output_sharding(self, x: torchax.tensor.Tensor):
        if self.enable_sp:
            token_num = x.shape[0]
            # NOTE(chengjiyao): make sure the sharded token_num is larger than TPU_SECOND_LAST_MINOR
            if token_num // self.tp_size < TPU_SECOND_LAST_MINOR:
                return None
        return self.output_sharding


class VllmQuantConfig:
    vllm_config: VllmConfig
    mesh: Mesh

    @classmethod
    def set_configs(cls, vllm_config: VllmConfig, mesh: Mesh):
        cls.vllm_config = vllm_config
        cls.mesh = mesh

    def get_linear_config(self, layer: LinearBase) -> VllmQuantLinearConfig:
        assert isinstance(layer, LinearBase)
        return VllmQuantLinearConfig(self.vllm_config, self.mesh, layer)

    def get_moe_config(self, layer: FusedMoE) -> FusedMoEConfig:
        assert isinstance(layer, FusedMoE)
        moe_config = layer.moe_config
        use_ep = self.vllm_config.parallel_config.enable_expert_parallel
        moe_config.moe_parallel_config.use_ep = use_ep
        return moe_config



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax.numpy as jnp
from absl.testing import absltest
from jax._src import test_util as jtu

from tpu_inference.layers.jax.rope_interface import apply_rope


class RopeInterfaceTest(jtu.JaxTestCase):

    def test_apply_rope_standard(self):
        seq_len = 2
        num_heads = 1
        head_dim = 8
        rope_theta = 10000.0

        inputs = jnp.ones((seq_len, num_heads, head_dim), dtype=jnp.float32)
        positions = jnp.arange(seq_len, dtype=jnp.float32)

        outputs = apply_rope(inputs,
                             positions=positions,
                             head_dim=head_dim,
                             rope_theta=rope_theta,
                             rope_proportion=1.0,
                             rope_input_ordering="split")

        # Check pos 0: should be identity (1.0)
        self.assertArraysAllClose(outputs[0], inputs[0])

        # Check pos 1: all elements should be rotated
        pos1_output = outputs[1, 0, :]
        for idx in range(head_dim):
            self.assertNotAlmostEqual(pos1_output[idx], 1.0, delta=1e-5)

    def test_apply_rope_partial_interleaved(self):
        seq_len = 2
        num_heads = 1
        head_dim = 8
        rope_theta = 10000.0
        rope_proportion = 0.5

        inputs = jnp.ones((seq_len, num_heads, head_dim), dtype=jnp.float32)
        positions = jnp.arange(seq_len, dtype=jnp.float32)

        outputs = apply_rope(inputs,
                             positions=positions,
                             head_dim=head_dim,
                             rope_theta=rope_theta,
                             rope_proportion=rope_proportion,
                             rope_input_ordering="interleaved")

        # In interleaved ordering:
        # pairs are (0,1), (2,3), (4,5), (6,7)
        # rope_angles = 2.
        # First 2 pairs are rotated: (0,1) and (2,3).
        # i.e., indices 0, 1, 2, 3.
        # Unrotated: 4, 5, 6, 7.

        pos1_output = outputs[1, 0, :]

        # Unrotated
        unrotated_indices = [4, 5, 6, 7]
        for idx in unrotated_indices:
            self.assertAlmostEqual(pos1_output[idx], 1.0, delta=1e-5)

        # Rotated
        rotated_indices = [0, 1, 2, 3]
        for idx in rotated_indices:
            self.assertNotAlmostEqual(pos1_output[idx], 1.0, delta=1e-5)


if __name__ == "__main__":
    absltest.main()



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the tpu-inference project

import os

# Disable CUDA-specific shared experts stream for TPU
# This prevents errors when trying to create CUDA streams on TPU hardware
# The issue was introduced by vllm-project/vllm#26440
os.environ["VLLM_DISABLE_SHARED_EXPERTS_STREAM"] = "1"
os.environ["LIBTPU_INIT_ARGS"] = "--xla_tpu_use_tc_device_shape_on_sc=true"

# Monkeypatch vLLM to avoid ImportError: cannot import name 'SamplingParams' from 'vllm'
# in vllm/v1/... submodules due to circular imports or lazy loading failures.
try:
    import vllm
    import vllm.sampling_params
    if not hasattr(vllm, "SamplingParams"):
        vllm.SamplingParams = vllm.sampling_params.SamplingParams
    if not hasattr(vllm, "SamplingType"):
        vllm.SamplingType = vllm.sampling_params.SamplingType
    if not hasattr(vllm, "SamplingStatus"):
        from vllm.sampling_params import RequestOutputKind
        vllm.RequestOutputKind = RequestOutputKind
except ImportError:
    pass


# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pytest
import torch
from vllm.pooling_params import PoolingParams
from vllm.sampling_params import SamplingParams
from vllm.v1.pool.metadata import PoolingMetadata, PoolingStates

from tpu_inference.runner.input_batch import CachedRequestState, InputBatch

# Default parameters for creating InputBatch instances in tests
MAX_NUM_REQS = 8
MAX_MODEL_LEN = 1024
MAX_NUM_BATCHED_TOKENS = 2048
VOCAB_SIZE = 32000
BLOCK_SIZES = [16]


@pytest.fixture
def input_batch():
    """Provides a clean InputBatch instance for each test."""
    return InputBatch(
        max_num_reqs=MAX_NUM_REQS,
        max_model_len=MAX_MODEL_LEN,
        max_num_batched_tokens=MAX_NUM_BATCHED_TOKENS,
        pin_memory=False,
        vocab_size=VOCAB_SIZE,
        block_sizes=BLOCK_SIZES,
        is_spec_decode=True,
    )


def create_dummy_request(req_id: str,
                         prompt_len: int = 10,
                         output_len: int = 5,
                         sampling_params: SamplingParams = None,
                         pooling_params: PoolingParams = None,
                         block_ids=None) -> CachedRequestState:
    """Helper function to create a CachedRequestState instance."""
    if sampling_params is None:
        sampling_params = SamplingParams(temperature=0.8, top_p=0.9, top_k=50)

    prompt_token_ids = list(range(prompt_len))
    output_token_ids = list(range(prompt_len, prompt_len + output_len))

    if block_ids is None:
        # Create dummy block ids based on length
        num_blocks = (prompt_len + output_len + BLOCK_SIZES[0] -
                      1) // BLOCK_SIZES[0]
        block_ids = [[i] for i in range(1, num_blocks + 1)]

    return CachedRequestState(
        req_id=req_id,
        prompt_token_ids=prompt_token_ids,
        mm_features=[],
        sampling_params=sampling_params,
        pooling_params=pooling_params,
        block_ids=block_ids,
        num_computed_tokens=0,
        lora_request=None,
        output_token_ids=output_token_ids,
    )


def test_initialization(input_batch: InputBatch):
    """Tests if the InputBatch is initialized with correct default values."""
    assert input_batch.max_num_reqs == MAX_NUM_REQS
    assert input_batch.num_reqs == 0
    assert len(input_batch.req_ids) == 0
    assert not input_batch.req_id_to_index
    assert input_batch.all_greedy
    assert input_batch.is_spec_decode


def test_add_request(input_batch: InputBatch):
    """Tests adding a single request to the batch."""
    req = create_dummy_request("req-1", prompt_len=20, output_len=4)
    input_batch.add_request(req)

    assert input_batch.num_reqs == 1
    assert "req-1" in input_batch.req_id_to_index
    assert input_batch.req_id_to_index["req-1"] == 0
    assert input_batch.req_ids == ["req-1"]
    assert len(input_batch.spec_decode_unsupported_reqs) == 0

    # Verify token data
    assert input_batch.num_prompt_tokens[0] == 20
    assert input_batch.num_tokens[0] == 24
    assert input_batch.num_tokens_no_spec[0] == 24
    expected_tokens = np.array(req.prompt_token_ids + req.output_token_ids)
    np.testing.assert_array_equal(input_batch.token_ids_cpu[0, :24],
                                  expected_tokens)

    # Verify sampling params
    assert input_batch.temperature_cpu[0] == 0.8
    assert input_batch.top_p_cpu[0] == 0.9
    assert input_batch.top_k_cpu[0] == 50


def test_add_multiple_requests(input_batch: InputBatch):
    """Tests adding multiple requests and checks their indices."""
    req1 = create_dummy_request("req-1")
    req2 = create_dummy_request("req-2")

    input_batch.add_request(req1)
    input_batch.add_request(req2)

    assert input_batch.num_reqs == 2
    assert input_batch.req_ids == ["req-1", "req-2"]
    assert input_batch.req_id_to_index["req-1"] == 0
    assert input_batch.req_id_to_index["req-2"] == 1
    assert input_batch.num_tokens[1] == len(req2.prompt_token_ids) + len(
        req2.output_token_ids)
    assert input_batch.num_tokens_no_spec[1] == len(
        req2.prompt_token_ids) + len(req2.output_token_ids)


def test_remove_request(input_batch: InputBatch):
    """Tests removing a request, which leaves a gap in the batch."""
    req1 = create_dummy_request("req-1")
    req2 = create_dummy_request("req-2")
    input_batch.add_request(req1)
    input_batch.add_request(req2)

    removed_index = input_batch.remove_request("req-1")

    assert removed_index == 0
    assert input_batch.num_reqs == 1
    assert "req-1" not in input_batch.req_id_to_index
    assert input_batch._req_ids[0] is None  # Slot is now empty
    assert input_batch._req_ids[1] == "req-2"
    assert "req-1" not in input_batch.greedy_reqs


def test_condense(input_batch: InputBatch):
    """Tests condensing the batch after removing requests."""
    reqs = [create_dummy_request(f"req-{i}") for i in range(4)]
    for req in reqs:
        input_batch.add_request(req)

    # Remove requests from the middle and start
    input_batch.remove_request("req-1")
    input_batch.remove_request("req-0")

    # Before condense: [None, None, "req-2", "req-3"]
    assert input_batch._req_ids[0] is None
    assert input_batch._req_ids[1] is None
    assert input_batch.num_reqs == 2

    # Condense should move req-2 and req-3 to the front
    empty_indices = sorted([0, 1], reverse=True)
    input_batch.condense(empty_indices)

    assert input_batch.num_reqs == 2
    assert len(input_batch.req_ids) == 2
    assert input_batch.req_ids == ["req-3", "req-2"]
    assert input_batch.req_id_to_index["req-2"] == 1
    assert input_batch.req_id_to_index["req-3"] == 0

    # Check if a property was moved correctly
    assert input_batch.num_tokens[0] == len(reqs[2].prompt_token_ids) + len(
        reqs[2].output_token_ids)
    assert input_batch.num_tokens_no_spec[0] == len(
        reqs[2].prompt_token_ids) + len(reqs[2].output_token_ids)


def test_swap_states(input_batch: InputBatch):
    """Tests swapping the states of two requests."""
    req1 = create_dummy_request("req-1", prompt_len=10, output_len=1)
    req2 = create_dummy_request("req-2",
                                prompt_len=20,
                                output_len=2,
                                sampling_params=SamplingParams(top_p=0.5))

    input_batch.add_request(req1)
    input_batch.add_request(req2)

    # Capture states before swap
    req1_tokens_before = input_batch.token_ids_cpu[0].copy()
    req2_tokens_before = input_batch.token_ids_cpu[1].copy()
    req1_top_p_before = input_batch.top_p_cpu[0]
    req2_top_p_before = input_batch.top_p_cpu[1]

    input_batch.swap_states(0, 1)

    # Check IDs and mappings
    assert input_batch.req_ids == ["req-2", "req-1"]
    assert input_batch.req_id_to_index["req-1"] == 1
    assert input_batch.req_id_to_index["req-2"] == 0

    # Check swapped data
    assert input_batch.top_p_cpu[0] == req2_top_p_before
    assert input_batch.top_p_cpu[1] == req1_top_p_before
    np.testing.assert_array_equal(input_batch.token_ids_cpu[0],
                                  req2_tokens_before)
    np.testing.assert_array_equal(input_batch.token_ids_cpu[1],
                                  req1_tokens_before)


def test_all_greedy_property(input_batch: InputBatch):
    """Tests the `all_greedy` property."""
    # Initially true
    assert input_batch.all_greedy

    # Add a greedy request, still true
    req_greedy = create_dummy_request(
        "req-g", sampling_params=SamplingParams(temperature=0.0))
    input_batch.add_request(req_greedy)
    assert input_batch.all_greedy

    # Manually add a random request for testing purposes
    input_batch.random_reqs.add("req-r")
    assert not input_batch.all_greedy

    # Remove it, should be true again
    input_batch.random_reqs.remove("req-r")
    assert input_batch.all_greedy


def test_get_pooling_metadata(input_batch: InputBatch):
    """Tests the get_pooling_metadata interface"""

    def states_eq(a: PoolingStates, b: PoolingStates):
        checks = [
            len(a.hidden_states_cache) == len(b.hidden_states_cache),
            all(
                torch.equal(x, y)
                for x, y in zip(a.hidden_states_cache, b.hidden_states_cache)),
        ]
        return all(checks)

    def meta_eq(a: PoolingMetadata, b: PoolingMetadata):
        assert a.prompt_token_ids is None and b.prompt_token_ids is None
        checks = [
            torch.equal(a.prompt_lens, b.prompt_lens),
            len(a.pooling_params) == len(b.pooling_params),
            len(a.pooling_states) == len(b.pooling_states),
            all(x == y for x, y in zip(a.pooling_params, b.pooling_params)),
            all(
                states_eq(x, y)
                for x, y in zip(a.pooling_states, b.pooling_states)),
            # ignore pooling cursor
        ]
        return all(checks)

    assert meta_eq(
        input_batch.get_pooling_metadata(),
        PoolingMetadata(
            prompt_lens=torch.tensor([], dtype=torch.int32),
            prompt_token_ids=None,
            pooling_params=[],
            pooling_states=[],
        ),
    ), "Initial value should be all empty"

    # Just some task value to pass assertion in PoolingMetadata.__post_init__
    pooling_param = PoolingParams(task="embed")
    pooling_state = PoolingStates()

    req_0 = create_dummy_request(
        "req-0",
        prompt_len=10,
        pooling_params=pooling_param,
    )
    input_batch.add_request(req_0)
    assert meta_eq(
        input_batch.get_pooling_metadata(),
        PoolingMetadata(
            prompt_lens=torch.tensor([10], dtype=torch.int32),
            prompt_token_ids=None,
            pooling_params=[pooling_param],
            pooling_states=[pooling_state],
        ),
    ), "Pooling states is populated by InputBatch object it self"

    input_batch.remove_request("req-0")
    assert meta_eq(
        input_batch.get_pooling_metadata(),
        PoolingMetadata(
            prompt_lens=torch.tensor([], dtype=torch.int32),
            prompt_token_ids=None,
            pooling_params=[],
            pooling_states=[],
        ),
    ), "After remove, back to empty state."
